import java.util.*;

/**
 *Besteht aus der 'obersten' Compound des Atoms.
 *
 *Ist atom1 gleich f(a, b(2)), dann ist die Klasse atom1 ein wrapper
 *um die Compound  f(a, b(2)).<p>
 *Siehe auch <a href="SatchParserBNF.html"><b>SatchParserBNF.html</b></a>.
 */
class Atom{

  private Compound compound;

  private int index = -1;

  Atom(FkSymbol fs, Vector t){
    compound = new Compound(fs,t);
  }

  Atom(Compound cmp){
    this.compound = cmp;
  }


  Compound getCompound(){
    return compound;
  }

  void bindTo(Term[] substitution){
    compound.bindTo(substitution);
  }

  void cloneYourValue(){
    compound.cloneYourValue();
  }

  Atom clonedValue(){
    return new Atom((Compound)compound.clonedValue());
  }

  void unbind(){
    compound.unbind(compound);
  }

  public String predicate(){
    return compound.predicate();
  }
  
  Vector internVars(Vector v){
    return compound.internVars(v);
  }

  /**
   *Dient der eindeutigen Indexierung gleich-prdikatiger Atome.
   *
   *Wenn "f\1" das <code>predicate</code> von atom1 und atom2 sind,
   *dann atom1.index() == atom2.index().
   */
  public int index(){
    if(index == -1)
      index = predicate().hashCode();

    return index;
  }

  public boolean equals(Object o){
    if(o instanceof Atom)
      return ((Atom) o).compound.equals(compound);
    else
      return false;
  }

  public String toString(){
    return compound.toString();
  }
}
    
