
/**
 *Eine wachsende Liste, die man sortieren kann.
 *
 *Die Liste hlt sich nach Benutzung von <code>sort</code> weiterhin selbst
 *sortiert, wenn man mittels <code>add</code> ein neues Atom einfgt.<br>
 *
 *Die Liste wird benutzt, um die Negation des jeweiligen Modells zu verein-
 *fachen. Da alle Atome ground sind, kann man einen reinen equals()-Vergleich
 *zwischen neuen (greren) und alten (kleineren) Modellen machen, um zu 
 *prfen, ob das ltere Modell Teilmenge des neuen Modells ist. 
 */
public final class AtomList {

  private Atom[] elems;
  private int elementCount;
  private boolean sorted = false;


  
  public AtomList(){ elems = new Atom[10]; }

  /* 
  public final void empty(){
   elems = new Atom[10]; elementCount = 0;
  }*/


  public final int size(){
    return elementCount;
  }

  /**
   *Ist die Liste sortiert, wirde jedes neue Atom in die bestehende Ordnung 
   *passend eingefgt.
   */
  public final void add(Atom at){


    if(elems.length == elementCount){
      Atom[] help = new Atom[elementCount+10];
      System.arraycopy(elems, 0, help, 0, elementCount);
      elems = help; help = null;
    }

    if(sorted){//Bei der zu erwartenden durchschnittlichen Lnge der Modelle
               //drfte das naiv implementierte Einsortieren schneller sein
               //als ein 'quicksortiges'.

      for(int i=0; i<elementCount;i++){
	if(at.index() <= elems[i].index()){

	  System.arraycopy(elems,i, elems, i+1, elementCount-i);
	  elems[i] = at;
	  elementCount++;
	  break;
	}

      }//for
    }
    else
      elems[elementCount++] = at;    
  }//add


  public Atom[] getAtoms(){
    Atom[] help = new Atom[elementCount];
    System.arraycopy(elems, 0, help, 0, elementCount);

    return help;
  }
  
  public final boolean isEmpty(){
    return elementCount == 0;
  }


  /**
   *Sortiert die Elemente der Liste nach Gre von Atom.predicate().hashcode().
   *
   *Da der Hashcode von Strings immer gleich ist, wenn auch die Strings 
   *gleich sind, kann man den Atom.predicate-String als Ordnungsindex benutzen.
   */
  public void sort(){
    if(sorted) return;

    quicksort(0, elementCount-1);
    sorted = true;
  }

  private void quicksort(int l, int r){
    int left = l; int right = r;

    int grenzwert = elems[(l+r)/2].index();
    
    while(left <= right){
      while(elems[left].index() < grenzwert)left++;
      while(elems[right].index() > grenzwert)right--;
      if(left<= right){

	if(left < right)
	  swap(left, right);
	
	left++;right--;
      }
    }

    if(l<right)quicksort(l, right);
    if(left<r)quicksort(left, r);
  }

  private void swap(int a, int b){

    Atom hilf = elems[a];
    elems[a] = elems[b];
    elems[b] = hilf;
  }

  /**
   *Prft ob die bergebene AtomList Teilmenge dieser AtomList ist.
   *
   *Dabei wird angenommen, da die bergebene AtomList kleiner ist als
   * 'this'! (Sonst knnten wir ja nicht auf *Teilmenge* prfen.)
   */
  public final boolean contains(AtomList alist){

    if(alist.isEmpty())
      return true;//unpraktisch, aber richtig

    if(alist.elementCount > elementCount)//Wenn alist mehr Atome als this hat,
      return false;                      //kann alist ja keine Teilmenge sein
    
    if(!alist.sorted) 
      alist.sort();

    int j=0;int start =0; int pos_in_this =0;

    alist_atoms: for(int i= 0; i<alist.elementCount;i++){

    this_atoms: for(;j<this.elementCount;j++){
      //um nicht immer bei 0 anzufangen
      pos_in_this++;


      //Dadurch, da die Listen sortiert sind, ist klar, da elems[i] nicht
      //in this enthalten sein kann, wenn man solange hat suchen mssen, da
      //der elems[i]-index nun kleiner ist als der von elems[j]. (Kleinere
      //j-Indizes werden nicht mehr kommen!)
      if(alist.elems[i].index() < this.elems[j].index()){
	//System.out.println("Modelle nicht gleich!!");
	return false;
      }
      //Sind die Indizes gleich, kann auf Gleichheit der Atome getestet
      //werden. 
      if(alist.elems[i].index() == this.elems[j].index()){
	if(start==0)start=j;
	if(alist.elems[i].equals(this.elems[j])){
	  //System.out.println("\tsind gleich!");
	  //Ist dies der Fall prfen wir noch alle anderen alist-Atome
	  //ab der Position an der der Index das erstemal gleich war (start).
	  j = start; start =0;
	  continue alist_atoms;
	}
	else//Sind die beiden Atome nicht gleich, und alle this-Atome 
	  //durchprobiert, dann sind wir alle Atome durchgegangen 
	  //(pos_in_this == elementCount), und haben ein alist-Atom, das 
	  //in this nicht enthalten ist -> return false
	  if( pos_in_this >= elementCount){
	    //System.out.println("Modelle nicht gleich!!");
	    return false;
	  }
	  else 
	    continue this_atoms;
      }//if index == index
      //System.out.println("elementCount: "+elementCount+" pos_in_this: "+pos_in_this);
      //Wenn die Atome nicht gleich sind, tun wir nichts, denn es kann 
      //ja mehrere mit gleichem index() geben.
      //
      //PROBLEM: wenn im Modell alle Atome den gleichen
      //index() haben, dann wird true zurckgegeben, auch wenn die Modelle 
      //in wahrheit verschieden sind!  
      //LSUNG: pos_in_this
      
      
    }//this_atoms
    }//alist_atoms
    
    //Alle alist-Atome erfolgreich verglichen
    return true;
  }//contains

  public final String toString(){
    StringBuffer buf = new StringBuffer();

    if(elementCount >= 1){
	buf.append("{ ");
      if(elems[0] != null)
	buf.append(elems[0].toString());
      else
	buf.append("null");

      for(int i= 1; i<elementCount;i++){
	if(elems[i] != null)
	  buf.append(", "+elems[i].toString());
	else
	  buf.append("null");
      }
    }
    else buf.append("[]");
    
    buf.append(" }");
    return buf.toString();
  }//toString


  public final String toSnarks(){
    StringBuffer buf = new StringBuffer();

    if(elementCount >= 1){
      if(elems[0] != null)
	buf.append(elems[0].toString());
      else
	buf.append("null");

      for(int i= 1; i<elementCount;i++){
	if(elems[i] != null)
	  buf.append(", "+elems[i].toString());
	else
	  buf.append("null");
      }
    }
    else buf.append("[]");
    
    return buf.toString();
  }//toString

    
}//class


















