import java.util.*;
/**
 *Eine Compound ist ein Prdikat mit einem mehreren oder keinen Termen als 
 *Argument.
 *
 *<p>
 *Siehe auch <a href="SatchParserBNF.html"><b>SatchParserBNF.html</b></a>.
 */
class Compound extends Term{
  //Variablen
  private FkSymbol fkSymbol;
  private Compound nonvar, klon;
  private Vector terme, vars; 
  private Var var;
  private Atom boundFrom = null;

  //sollen nicht-private bleiben!
  Vector upperVars_vec = null; 
  boolean[] todo = null;


  //Konstruktoren
  Compound(FkSymbol fs, Vector t){
    fkSymbol = fs; terme = t;
  }

  Compound(FkSymbol fs){
    fkSymbol = fs; terme = new Vector();
  }
  

  //Instanz-Methoden
 
  public boolean equals(Object ob){
    if(ob instanceof Compound){
      Compound other = (Compound) ob;
      if(!other.predicate().equals(this.predicate())) return false;

      for(int i=terme.size()-1;i>-1;i--){
	Term t1 =((Term) terme.elementAt(i)).value(); 
	Term t2 = ((Term) other.terme.elementAt(i)).value();

	if(! t1.equals(t2))
	  return false;
      }
      return true;
    }
    else return false;
  }

  /**
   *Matcht einen Term, wenn er eine Compound oder eine Var ist.
   *
   *Der anfragende Term muss der sein, an dem .matches() ausgefhrt wird.
   *D.h. was in der Datenbank steht muss an das anfragende Atom bergeben 
   *werden.
   */
  public boolean matches(Term t, Compound comp){
    if(t instanceof Compound){
      nonvar = (Compound) t;
      
      if(nonvar.terme == null)//ist eine Konstante
	return this.fkSymbol.equals(nonvar.fkSymbol);

      else if(this.fkSymbol.equals(nonvar.fkSymbol) &
	      terme.size() == nonvar.terme.size()){
	
	for(int i = 0;i< terme.size();i++) {
	  Term myterm  = (Term) terme.elementAt(i);
	  Term otherTerm = (Term) nonvar.terme.elementAt(i);
	  if(! myterm.matches(otherTerm, comp))
	    return false;
	}
	return true;
      }
    }
    if(t instanceof Var)
      return t.matches(this, comp);

    else return false;//t weder Var noch Compound
  }

  /**
   *Gibt Prdikatsymbol und Stelligkeit aus.
   *
   *Beispiel:<p>
   *Die Compound <code>p(a,f(88,9),J)</code> wrde <code>p/3</code> 
   *zurckgeben.
   */
  String predicate(){
    return fkSymbol+"/"+terme.size();
  }  

  Term value(){
    return this;
  }

  /**
   *Im ersten Schritt mssen alle Values geklont werden. Dann wird die Funktion
   *<code>clonedValue</code> aufgerufen, die den geklonten Wert zurckgibt.<br>
   *<br><strong>Nach einem Aufruf von cloneYourValue muss die Funktion <code>unbind
   *</code> aufgerufen werden!</code><br>Geschieht dies nicht bleiben die gematchten
   * Variablen gebunden, und die DataBase verliert ihren Wert.
   */
  void cloneYourValue(){

    if(terme.size() ==0){//o.println("\t"+fkSymbol); //mal ausprobieren! erstaunlich!
      klon = this; //Sharing dieser Instanz
    }
    else{
      //Shallow copie des Terme-Vectors (die Position der Terme bleibt erhalten)
      Vector t2 = (Vector) terme.clone();//o.println("t2 frisch geklont "+t2);
      
      //klone jeden Term
      for(int i=t2.size()-1;i>-1;i--){
	((Term) t2.elementAt(i)).cloneYourValue();//o.println("ok3a");
      }

      //ersetzen jeden Term durch value-Klon    
      for(int i=t2.size()-1;i>-1;i--){
	t2.setElementAt(((Term) t2.elementAt(i)).clonedValue(), i);
      }
      
      klon =  new Compound(fkSymbol, t2);
    }
    //hier ein unbind zu machen wre verfrht:(this knnte eine sub-Compund sein)

  }//cloneYourValue


  public Term clonedValue(){
    if(klon != null)
      return klon;
    else
      throw new NoSuchElementException("No Clone generated!");
  }

  void bindTo(Term[] werte){

    for(int i=0; i<werte.length; i++)
      if(werte[i] != null)
	((Var) terme.elementAt(i)).bindTo(werte[i], this);
  }


  /**
   *Diese Methode entbindet nur die Variablen in dieser Compound,
   *die von der bergebenen Compound gebunden wurden.
   *
   *<BR>Sonst kommt es bei folgender Eingabe zu einem bsen Fehler bei 
   *folgender Spezifikation
   *<UL>--->p(a).
   *<BR>--->p(b).
   *<BR>--->r(c).
   *<BR>r(X), p(X) ---> foo(X).</UL>
   *kam als Modell p(a), p(b), <B>r(b)</B> und <B>foo(b)</B>, obwohl die letzte
   *Klausel eigentlich unerfllbar ist! Das lag daran, da beim
   *Backtracken nicht nur p(a) sondern auch r(a) von der Bindung gelst
   *wurde.
   *<BR>Wie bereits erwhnt, steht ja an jeder Stelle der klausel die
   *selbe Variablen-Instanz. Beim Matching wurde im ersten Schritt das X in
   *r(X) an c gebunden. Dann wurde versucht p(c) zu beweisen, was nicht gelang,
   *da p(c) nicht zu den Axiomen gehrt (d.h. nicht in der Database enthalten
   *war). Da der Versuch des Beweisens scheiterte, wurde ein neuer Versuch gestartet:
   *p(c) wurde von der konstanten c gelst, und es wurde versucht p(X)
   *zu beweisen. <I>Aber genau das war der Fehler!</I> Das X in p(X) wurde
   *von r(X) an c gebunden, und deshalb darf auch nur r(c) die Variable X von
   *c lsen.
   *<BR>Der oben erwhnte zweite Beweisversuch gelang, p(X) wurde an b
   *gebunden, und damit auch r(X).
   *<BR>Siehe auch in der <A HREF="code.html#killer">Dokumentation</A>.
   */
  void unbind(Compound comp){

    for(int i=terme.size()-1;i>-1;i--)
      ((Term) terme.elementAt(i)).unbind(comp);
    
    klon = null;
  }



  /**
   *Gibt einen Array von Variablen-Bindungen zurck.
   *Genau genommen ist das Array ein Abbild der Position und der Werte der 
   *Variablen in der Compound. An jeder Stelle an der eine gebundene Variable
   *steht, steht im zurckgebenen Array der Wert der Variablen.<p>
   *
   *Es wird entweder ein Array oder null zurckgegeben. <strong>Das Array hat an 
   *den Positionen, an denen keine Variablen-Werte stehen ebenfalls null stehen.
   *</strong>
   */
  public Term[] getSubstitutions(Compound comp){

    if(comp.matches(this, this)){ 

      Term[] substitution = comp.getVarsCopy();
      comp.unbind(this); this.unbind(this);
      return substitution;
    }
    else{

      comp.unbind(this); this.unbind(this);
      return null;
    }
  }//getSubstitutions


  public Term[] getSubstitutions(Atom at){
    return getSubstitutions(at.getCompound());
  }


  private Term[] getVarsCopy(){

    Term[] copy = new Term[terme.size()];
    this.cloneYourValue();

    for(int i=terme.size()-1;i>-1;i--)
      if(terme.elementAt(i) instanceof Var && ((Var) terme.elementAt(i)).isBound())
	copy[i] = ((Var) terme.elementAt(i)).clonedValue();
    
    return copy;
  }


  public String toString(){
    if(terme.size() ==0)
      return fkSymbol.toString();

    else{
      String o;
      o = fkSymbol.toString(); o += "(";

      for(int i=0; i< terme.size();i++){
	o += ((Term) terme.elementAt(i)).toString();
	if(i<terme.size()-1)
	  o += ", ";
      }
      return o+=")";
    }
  }//toString

  /**
   *Diese Methode wird vom Parser nach dem Parsen aufgerufen, und stellt 
   *sicher, da an jeder Stelle der Compound, an der zwei gleichlautende 
   *Variablen stehen, das selbe Var-Objekt steht.
   *
   *Dabei wird ein Vector angelegt (falls keiner bergeben wurde), in den 
   *alle Variablen-Objekte hineingeschrieben werden, die bisher vorkamen.
   *Dieser Vector wird an alle in dieser Compound vorkommenden Compounds 
   *bergeben, damit alle Variablen, die in den untergeordneten Compounds
   *vorkommen, durch die in dieser Compound vorkommenden Variablen-Objekte 
   *ersetzt werden. 
   *Damit auch alle Variablen die zwar nicht in dieser, aber in zwei oder 
   *mehreren untergeordneten Compounds vorkommen ersetzt werden knnen, wird 
   *der bergebene Vector wieder an die aufrufende Compound zurckgegeben.
   */
  private void internVars(){

    Compound nonVars[] = new Compound[terme.size()];int nvPos =0;
 
    boolean someVar_left = true;
      
    while(someVar_left){
      someVar_left = false;

      //todo wird ininternVars(Vector var_vec) 
      //initialisiert und hat die Lnge terme.size()
      for(int i=0; i<todo.length;i++){    
	if(todo[i]){                           //wenn zu erledigen

	  if(terme.elementAt(i) instanceof Var){//und Variable
	    if(var == null){
	      var = (Var) terme.elementAt(i);
	      todo[i]=false;

	      if(!var.getName().startsWith("_")){ //Schlielich sollen zwei '_'
		vars.addElement(var);            //verschiedene Variablen sein.

	      }
	      else var = null;
	      
	    }
	    else if(var.equals(terme.elementAt(i))){//ersetze
	      terme.setElementAt(var,i);          //wenn gleich
	      todo[i]=false;

	    }
	    else{ someVar_left = true; //Var ist nicht gleich

	    }
	  }//if(terme.elementAt(i) instanceof Var)
	  else if(terme.elementAt(i) instanceof Compound){
	    nonVars[nvPos] = (Compound) terme.elementAt(i);
	    nvPos++;
	  }
	  else{ todo[i]=false; //wenn nicht Var od Compound

	  }
	}
      }//for(todo.length...
      var = null; //fr diese Var wurden alle mglichen Ersetzungen gemacht.
    }//while
    //Alle gleichen Variablen sollten jetzt behandelt sein.
      
    //Auch alle Compounds sollten 'unifiziert' werden
    for(int i =0;i<nvPos; i++){

      vars = nonVars[i].internVars(vars);

    }
  }//internVars()...

  /**
   * Siehe <code>internVars()</code>.
   *Diese Methode ersetzt alle in der Compound enthaltenen Variablen durch die
   *bergebenen, <strong>falls</strong> sie gleich sind. Von hier aus wird dann
   * <code>internVars()</code> aufgerufen.
   **/
  Vector internVars(Vector var_vec){
    
    upperVars_vec = var_vec;
    vars = new Vector();

    todo = new boolean[terme.size()];
    for(int i = todo.length-1; i>-1; i--)todo[i]=true;
    
    //fr jede bergebene Variable ...
    for(int j=upperVars_vec.size()-1;j>-1;j--){
      var = (Var) upperVars_vec.elementAt(j);

      vars.addElement(var);
      
      //gehe alle Terme durch...
      for(int i=0; i<todo.length;i++){
	if(todo[i]){                           //wenn zu erledigen
	  
	  if(terme.elementAt(i) instanceof Var){//und Variable
	    
	    if(var.equals(terme.elementAt(i))){//ersetze
	      terme.setElementAt(var,i);       //wenn gleich
	      todo[i]=false;

	    }
	  }
	}//if(todo...
      }//for(todo.length...
    }//for(upperVars_Vec.size()
  

    internVars();

    return vars;
  }//internVars
      
}//class Compound
