import java.util.*;
import java.io.*;
/**
 *Satchmo ist ein von Bry/Manthey entwickeltes Modellgenerierungsverfahren. 
 *
 *Die erste Satchmoversion war ein Prologprogramm. Dies ist die
 *Javaimplementierung von Satchmo. Dieses Programm ist ein Kommandozeilentool.
 *Der aufruf auf der Kommandozeile lautet:<p>
 *  java Satchmo [options] file [files]<p>
 *Es gibt folgende Optionen:<p>
 *-mm<spacer type=horizontal size=8>generates only Minimal Models.<br>
 *<spacer type=horizontal size=12>Option -cs is therefore included in this option<br>
 *-cs<spacer type=horizontal size=8>generates each model only once (ComplementSplitting).<br>
 *-maxmodels zahl<spacer type=horizontal size=8>generates only zahl models<br>
 *-trace<spacer type=horizontal size=8>generates trace-files for SNARKS<br>
 *-verbose<spacer type=horizontal size=8>print some information about the momentaneous <br>
 * state of inference, whereas means<br>
 * <spacer type=horizontal size=8>Neg_Atoms: negated complents when cs is turned on<br>
 * <spacer type=horizontal size=8>Neg_Models:   negated models when mm is turned on<br>
 * <spacer type=horizontal size=8>vI:           heads of clauses, which are violated instances<br>
 *<spacer type=horizontal size=16>('null' means 'empty head', i.e. a negation.)<br>
 * <spacer type=horizontal size=8>assert/retract: assert/retract an atom to/from database<br>
 */
class Satchmo{

  private Vector clauseBase = null;
  private DoubleQueue violatedInstances = null;
  private DataBase db;

  private boolean cs = false, mm = false, verbose = false, trace = false, 
    maxmodels =false;

  private PrintWriter traceFileStream = null;
  private Negations negations;

  private SatchParser parser;
  private PrintStream o = System.out;
  private String filename = null;
  private int modelCount = 0, maxModelNumber = 0;

  private static final String helpString = 
  "\nusage: java Satchmo [options] file [files]"+
  "\noptions are:"+
  "\n     -mm\tgenerates only Minimal Models."+
  "\n\t     \tOption -cs is therefore included in this option"+
  "\n\n     -cs\tgenerates each model only once (ComplementSplitting)."+
  "\n\n     -maxmodels <n>\tgenerates only n models"+
  "\n\n     -trace\tgenerates trace-files for SNARKS"+
  "\n\n     -verbose   print some information about the momentaneous "+
  "\n\t        state of inference, whereas means"+
  "\n\t        Neg_Atoms:    negated complents when cs is turned on"+
  "\n\t        Neg_Models:   negated models when mm is turned on"+
  "\n\t        vI:           heads of clauses, which are violated instances"+
  "\n\t                      ('null' means 'empty head', i.e. a negation.)"+
  "\n\t        assert/retract assert/retract an atom to/from database";


  /**
   *Konstruktor wird nur beim ersten Initialisieren benutzt. 
   *
   *Bei weitere Initialisierungen ReInit benutzen!
   */
  Satchmo(String file){

    try{
    java.io.FileInputStream fis = new java.io.FileInputStream(file);

      parser = new SatchParser(fis);

      filename = file;

      clauseBase = parser.Read();

    }catch(Throwable e){System.out.println("\nerror encountered:\n");
                        o.println(e);
				 System.exit(0);}

  }

  /**
   *Wird bentigt um Satchmo und den Parser zu reinitialisieren.
   *
   *Die von JavaCC generierten Parser sind aus Effizienzgrnden static, und
   *knnen deshalb nicht neu Instanziert werden. Man mu die <code>ReInit()
   *</code>-Methode des Parsers benutzen; das hat zur Folge, da man auch 
   *<code>Satchmo</code> nicht einfach neu instatieren darf. Benutzen Sie 
   *stattdessen diese Methode!
   */
  public void ReInit(String file){
    
    modelCount = 0; //Modelle-Zhler zurcksetzen
    try{
    java.io.FileInputStream fis = new java.io.FileInputStream(file);

      parser.ReInit(fis);
      filename = file;
      clauseBase = parser.Read();

    }catch(java.io.FileNotFoundException fnf){o.println("File: "+file+" not found!");}
    catch(Throwable e){System.out.println("\n\terror encountered:\n");
                        o.println(e);
    }
  }//ReInit


  public static void main(String args[]){

    if(args.length ==0){
      //kleine Hilfe ausgeben
      System.out.println(Satchmo.helpString);
      System.exit(0);
    }

    Satchmo satchmo = null;

    for(int i=0; i<args.length;i++){

      if(satchmo == null){
	boolean mm_=false; boolean cs_ = false, verbose_ =false, 
			     trace_ = false, maxmodels_ = false;
	int maxModelNumber_=0;
	
      loop: while(true){
	
	if(args[i].equals("-mm")){
	  mm_ = true; 
	  i++; 
	  continue loop;
	}
	else if(args[i].equals("-cs")){
	  cs_ = true; 
	  mm_ = false; 
	  i++; 
	  continue loop;
	}
	else if(args[i].equals("-verbose")){
	  verbose_ = true;
	  i++;
	  continue loop;
	}
	else if(args[i].equals("-trace")){
	  trace_ = true;
	  i++;
	  continue loop;
	}
	else if(args[i].equals("-help")){
	  System.out.println(Satchmo.helpString);
	  System.exit(0);
	}
	else if(args[i].equals("-maxmodels")){
	  maxmodels_ = true;
	  i++;

	  try{

	    maxModelNumber_ = Integer.valueOf(args[i]).intValue();

	  }catch(NumberFormatException nfe){System.out.println(args[i]+" is "+
	  "not a number!\n\tIf you want to use 'maxmodels' you have to give"+
	  " a number as argument."); System.exit(0);}

	  i++;
	  continue loop;
	}

	System.out.println("\n\n__________Saturate: "+args[i]+"__________");
	
	satchmo = new Satchmo(args[i]);
	satchmo.mm = mm_; satchmo.cs = cs_; 
	satchmo.verbose = verbose_; satchmo.trace = trace_; 
	satchmo.maxmodels = maxmodels_; satchmo.maxModelNumber = maxModelNumber_;

	satchmo.refute();
	break loop;
      }//loop: while...
      }//if satchmo == null
      else{
   
	System.out.println("\n\n__________Saturate: "+args[i]+"__________");
	long start = System.currentTimeMillis();

      satchmo.ReInit(args[i]);
      satchmo.refute();
      if(satchmo.verbose) System.out.println("total: "+(System.currentTimeMillis()-start));

      }//else
    }//for
  }//main


  /**
   *Nachdem die Spezifikation geladen ist, wird <code>refute()</code> 
   *aufgerufen um den Modellgenerierungsprozess zu starten.
   */
  void refute(){

    db = new DataBase();
    violatedInstances = new DoubleQueue();
    FileOutputStream fos=null;

    if(cs|mm) negations = new Negations();

    if(trace){// Trace-datei ffnen.
      try{
	//Datei ffnen
	fos = new FileOutputStream(filename+".trace");
	//Stream ffnen (wir flushen-> true)
	traceFileStream = new PrintWriter(fos, true);

      }catch( IOException ioe){o.println("\t!! Could not write "
                                    +filename+".trace : "+ioe);}
    }
    //
    //Hier kommt der eigentliche Beweisprozess in Gang.
    satisfiable();
    if(modelCount ==0)o.println("\nNo model found!");

	
    if(trace){// Trace-datei schreiben.
      try{
	//Schlieen
	traceFileStream.close(); fos.close();
	o.println("\nTrace written to "+filename+".trace\n");

	}catch( IOException ioe){o.println("\t!! Could not write "
                                    +filename+".trace : "+ioe);}
    }

  }//refute
    
  /**
   *Entspricht satisfiable in <a href="satchmo.prolog"><b><code>satchmo.prolog</code></b></a>.
   */
  void satisfiable(){ 

    //Start: violated Instances (vI) suchen, falls keine mehr da sind
    if(violatedInstances.isEmpty()){
      
      for(int i =0;i< clauseBase.size();i++){
	Clause c = (Clause) clauseBase.elementAt(i);

	//Der Nuller dient als Positionsmarker in der Klausel;
	//ist c eine vI, wird c nach 'violatedInstances' geschrieben
	// --> siehe violatedInstance
	violatedInstance(c,0);
      }
      
      //Die Komplemente fr cs einfgen (die Komplement-Atome wurden in 
      //satisfy_all in die negations eingefgt).
      if(cs|mm){
	  Vector negAtoms = negations.getAtoms();

	  Vector appliableComplements = db.selectContainedAtoms(negAtoms); 
	  for(int j= 0;j<appliableComplements.size();j++){
	      
	    if(trace) 
	      violatedInstances.enqueue(appliableComplements.elementAt(j)
					+" ---> false"); 
	      else 
		violatedInstances.enqueue(null);
	  }
	//Man knnte meinen , da hier ein Fehler liegt:
	//In sf_all werden ALLE Atome der Disjunktion in die negations
	//eingefgt, und hier wird fr ALLE Atome in den negations
	//'null' (negation) in die vI geschrieben. Aber wir wollen doch
	//nicht fr das erste Atom der Disjunktion auch ein 'null' haben! (Oh!)
	//
	//Haben wir auch nicht! Bevor satisfy wieder aufgerufen wird, wird das
	//erste Atom der Disjunktion 'asserted', und dabei wird es ja auch
	//aus den 'negations' rausgenommen. (Deshlab ist der cs-Code auch so 
	//komisch zwischen den Methoden zersplittert.) Alles klar?
      }
      
      //Sollte eines der bisher gefundenen Modelle momentan erfllt sein,
      //so mu es fr minimal modell negiert werden.
      if(mm){
	  Vector negModels = negations.getModels();

	  Vector appliableModels = db.selectContainedModels(negModels); 
	  for(int j= 0;j<appliableModels.size();j++){
	      
	      if(trace) 
		violatedInstances.enqueue(((AtomList) appliableModels.elementAt(j)).toSnarks()+" ---> false"); 
	      else
		violatedInstances.enqueue(null);
	  }    
      }
    
      //wenns keine vI mehr gibt, haben wir ein Modell
      if(violatedInstances.isEmpty()){

	modelCount++;
	o.println("\nmodel "+modelCount+":"); 
	db.print();
	
	//Abbrechen, wenn maxmodel Maximum erreicht ist
	if(maxmodels && (maxModelNumber <= modelCount))
	  System.exit(0);
	
	//das Modell negieren, wenn minimal model erwnscht ist
	if(mm){
	  AtomList  atomList = db.getAtoms2();
	  negations.add(atomList);
	}

	if(trace) traceFileStream.println("model");

	return; //weil wir ja noch weitere Modelle suchen wollen
      }
    }//if(violatedInstances.isEmpty()
    
    while(!violatedInstances.isEmpty())//Hier die fairness
      satisfy_all();
	 
  }//satisfiable


 /**
   *Entspricht satisfy_all in <a href="satchmo.prolog"><b><code>satchmo.prolog</code></b></a>.
   */
  void satisfy_all(){ 

    if(verbose) o.println((negations==null ? "" :"\n"+negations.toString()) 
			  +"\nvI: ["+violatedInstances+"]");

    //Ist das Objekt ein String, dann ist der Head 'null' (der String ist
    //nur ein Abbild der angewandten Regelinstanz). Dieser String ist sonst
    //in den Heads als 'instanceImage' enthalten. 
    //MERKE: der String ist nur enthalten wenn trace == true!!
    Object obj = violatedInstances.dequeue();     Head head;

    if(verbose)o.println("dequeue: "+(obj==null ? "false" : obj.toString()));

    if(obj == null || obj instanceof String ){//Negation in  der Spezifikation

      head=null;

	//Tracefile ber Negation informieren
	if(trace) traceFileStream.println("applied_rule_instances(["+
				     obj+"])\n"+
				     "new_successor_node\n"+
				     "new_atoms([false])\n"+
				     "closed\n"+
				     "back");

      return;
    }

    //ist es kein String, dann ist es ein Head.
    head = (Head) obj;
    Vector headAtoms = head.getAtoms();

    if(db.containsOneOf(headAtoms)){ //Wenn die db ein Atom des Heads enthlt,
      if(verbose) o.println("\t"+headAtoms+" already in DataBase");
      satisfiable();             // ist diese Klausel keine vI mehr; ->weiter

    }
    else{//Hier erledigen wir die Disjunktion der Head-Atome

      if(cs|mm){//Alle Head-Atome negieren, falls cs angeschaltet

	for(int i = 0;i<headAtoms.size();i++)
	  negations.add((Atom)headAtoms.elementAt(i));
      }
      //Tracefile mit Information ber die angewandte Instanz versehen
      if(trace)
	traceFileStream.println("applied_rule_instances(["+
			   head.applied_instance_image()+"])");


      //Die Atome nacheinander annehemen und wieder zurcknehmen
      for(int i = 0;i<headAtoms.size();i++){
	Atom a= (Atom)headAtoms.elementAt(i);

	if(cs|mm) negations.removeAtom(a);

	if(verbose) o.println("\t\tassert: "+a);

	if(trace) traceFileStream.println("new_successor_node\n"+
				     "new_atoms(["+a+"])");
if(db.containsOneOf(headAtoms))o.println("Huhu, Pech gehabt!\t Atom: "+a+" HeadAtoms: "+headAtoms);
	db.assert(a);

	satisfiable();
	db.retract(a); //Beim Zurckkehren kann man auf jeden Fall das 
	               //'angenommene' (asserted) Atom wieder entfernen.
	               if(verbose) o.println("\t\tretract: "+a);
		       if(trace) traceFileStream.println("back");
      }
    }
    
  }//satisfy_all



  /**
   *Es wird versucht den Rumpf (Body) der bergebene Klausel aus dem 
   *bisherigen Bestand an bewiesenen (oder angenommenen) Klauseln in der 
   *<code>DataBase</code> zu beweisen.
   *
   *Ist der Rumpf beweisbar, wird geprft, ob eine Element des Kopfes 
   *bereits in der <code>DataBase</code> enthalten ist. Ist dies der 
   *Fall, so ist die Klausel keine verletzte Instanz einer Regel.
   *Der Positionsmarker <code>position</code> dient der eindeutigen
   *Bestimmung des zu prfenden Atoms der Klausel. Zur einfachen 
   *berprfung, ob die Klausel eine verletzte Instanz (vI) einer Regel 
   *ist, ist folgender Aufruf genau der richtige:<p>
   *violatedInstance(klausel_xyz, 0);<p>
   *Warum benutze ich keine while-Schleife, sondern Rekursion?<p>
   *Um Backtracken zu knnen! In einer Schleife kann ich das nicht.
   *<p>
   *Ist die Klausel eine vI, so findet sie sich mit einer (durch
   *die Spezifikation) bestimmten Belgung in den violatedInstances wieder.
   *<p>
   *Entspricht violated_instance in <a href="satchmo.prolog"><b><code>satchmo.prolog</code></b></a>.
   */
  private void violatedInstance(Clause c,int position){

      Head h = c.getHead();

    if(c.isTrue()){

      //Auch die leere Ableitung erlauben! (head & body null) 
      if(h== null){

	//Da man einem Nullpointer keinen String bergeben kann, behelfe
	//ich mir indem ich den String statt dem Nullpointer bergebe
	if(trace) 
	  violatedInstances.enqueue(c.toString()); 
	else
	  violatedInstances.enqueue(h);
      }
      else if(!db.containsOneOf(h.getAtoms())){
     
	violatedInstances.enqueue(h);//kann ja keine Variablen enthalten, da
	                             //c.isTure bedeutet c.body == null!
      
	//Fr die Ausgabe in die Trace-Datei
	if(trace) h.set_applied_instance_image(c.toString());
      }

      return;
    }//das else spar ich mir...

    Vector bodyAtoms = c.getBody().getAtoms();
  
    if(bodyAtoms.size() == position){//Body bewiesen, da position in Klausel 
                                     //grer als die Anzahl der Atome im Body

      if(h== null ){//h==null wenn 'x--->.'
	  
	//Da man einem NullPointer keinen String bergeben kann, behelfe
	//ich mir indem ich den String statt dem Nullpointer bergebe.
	if(trace) 
	  violatedInstances.enqueue(c.toString()); 
	else
	  violatedInstances.enqueue(h);

	if(verbose)o.println("\tClause: "+c);
      }
      else if(!db.containsOneOf(h.getAtoms())){//Ist vI wenn die db kein Atom 
	Head copy = h.copy();                  //des Heads enthlt
	violatedInstances.enqueue(copy);

	//Fr die Ausgabe in die Trace-Datei
	if(trace) copy.set_applied_instance_image(c.toString());
	if(verbose) o.println("applied Instance: "+c);
      }

    }
    else{//prfen, ob die Atome des Body matchen
	
      Atom atom = (Atom) bodyAtoms.elementAt(position);

      //die db gibt eine Liste aller mglichen Subtitutionen fr das Atom zurck

      Vector substitutions = db.getSubstList(atom);

      //Hier wird das Backtracking erledigt (for-Schleife)
      for(int i=0;i<substitutions.size();i++){//size == 0, wenn keine Subst gef.
	Term[] subst = (Term[]) substitutions.elementAt(i);
	
	//Die Variablen des Atoms werden mit den Werten aus der 
	atom.bindTo(subst);                  //Substitution belegt

	violatedInstance(c, position+1);//Mit dieser Belegung das nchste Atom testen

	atom.unbind();//Die Variablen von der Bindung befreien

      }//for
    }
  }//violatedInstance


}//class





