import java.util.*;

/**
 *Eine Variable wird immer vom anfragenden <code>Term</code> 
 *gematcht, auer sie ist bereits gebunden.
 *
 *Der Wert (d.h. das gebundene Objekt, falls gebunden) wird durch
 *die <code>value</code> Methode ermittelt. Wenn die Variable
 *gebunden ist, ruft <code>value</code> die <code>value</code> 
 *Methode des an die Variable gebundenen Terms auf; ansonsten 
 *wird die Variable selbst zurckgegeben.<p>
 *Siehe auch <a href="SatchParserBNF.html"><b>SatchParserBNF.html</b></a>.
 */
class Var extends Term{

  private String str;
  private Term boundBy;//Der Term, dessen Wert die Var hat.
  private Term klon;
  private Compound boundIn; //Die Compound in der die Var gebunden ist

  Var(String s){
    str = s;
  }

  String getName(){
    return str;
  }

  void bindTo(Term t, Compound comp){
    boundBy = t;
    boundIn = comp;
  }
  
  boolean isBound(){
    return boundBy != null;
  }

  public boolean equals(Object o){
    if(o instanceof Var){ 
      Var othervar = (Var) o;
      return str.equals(othervar.str);
    }
    else return false;
  }

  /**
   *Prft, ob der bergbener Term gematcht werden kann. 
   *
   *Ist die Variable an einen anderen Term gebunden, wird versucht den 
   *bergebenen Term gegen den anfgenden zu matchen.
   *Ist die Variable ungebunden, erhlt die Variable den Wert des bergebenen
   *Terms.
   */
  public boolean matches(Term t, Compound comp){

    if(boundBy != null){//wenn gebunden

      return boundBy.matches(t, comp);
    }
    else{              // wenn frei 
      if(t != this){
	boundBy = t;
	boundIn = comp;
      }
      return true;
    }
  }

  void unbind(Compound comp){
    if(boundBy != null)//wenn gebunden
      boundBy.unbind(comp);

    if(boundIn == comp){
      boundBy = null; klon = null;
    }
  }

  

  Term value(){
    if(boundBy != null)        //wenn gebunden
      return boundBy.value();
    else                       //wenn frei
      return this;
  }

  void cloneYourValue(){
    if(boundBy != null){        //wenn gebunden
      
      boundBy.cloneYourValue();
      
      klon = boundBy.value();
    }
    else                       //wenn frei
      klon =  new Var(str);    //klone dich selbst
  } 

  public Term clonedValue(){
    if(klon != null)
      return klon;
    else
      throw new NoSuchElementException();
  }


  public String toString(){
    if(boundBy != null)
      return boundBy.toString();
    else{
      return str;
    }
  }

}//class Var


